# BrainGenix-NES
# AGPLv3


class Configuration():

    def __init__(self):
        
        # Create Attributes
        self.Name:str = None
        self.SourceCompartment = None
        self.DestinationCompartment = None
        self.Conductance_nS:float = None
        self.TimeConstant_ms:float = None
        self.ReceptorLocation_um:list = None
    



# BrainGenix-NES
# AGPLv3

import json

from . import Configuration

from BrainGenix.NES.Client import RequestHandler

import BrainGenix.LibUtils.GetID
import BrainGenix.LibUtils.ConfigCheck

class BS:

    def __init__(self, _Configuration:Configuration, _RequestHandler:RequestHandler, _SimulationID:int):
        # Create Attributes
        self.Name = _Configuration.Name
        self.RequestHandler = _RequestHandler

        # Run Configuration Check
        BrainGenix.LibUtils.ConfigCheck.ConfigCheck(_Configuration)

        # Create Box On Server
        ShapeID = BrainGenix.LibUtils.GetID.GetID(_Configuration.Shape)
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Compartment/BS/Create?SimulationID={_SimulationID}&Name={_Configuration.Name}&ShapeID={ShapeID}&MembranePotential_mV={_Configuration.MembranePotential_mV}&SpikeThreshold_mV={_Configuration.SpikeThreshold_mV}&DecayTime_ms={_Configuration.DecayTime_ms}&RestingPotential_mV={_Configuration.RestingPotential_mV}&AfterHyperpolarizationAmplitude_mV={_Configuration.AfterHyperpolarizationAmplitude_mV}")
        assert(Response != None)
        self.ID = Response["CompartmentID"]


# BrainGenix-NES
# AGPLv3


def GetID(_Object):

    ShapeID = None
    if (type(_Object) == str):
        ShapeID = str(_Object)
    elif (type(_Object) == int):
        ShapeID = str(_Object)
    else:
        ShapeID = str(_Object.ID)

    return ShapeID


# BrainGenix-NES
# AGPLv3

import json

from . import Configuration

from BrainGenix.NES.Client import RequestHandler

import BrainGenix.LibUtils.GetID
import BrainGenix.LibUtils.ConfigCheck


class PatchClampDAC:

    def __init__(self, _Configuration:Configuration, _RequestHandler:RequestHandler, _SimulationID:int):
        # Create Attributes
        self.Name =  _Configuration.Name
        self.RequestHandler = _RequestHandler
        self.SimulationID = _SimulationID

        # Run Configuration Check
        BrainGenix.LibUtils.ConfigCheck.ConfigCheck(_Configuration)

        # Create On Server
        ClampLocation = json.dumps(_Configuration.ClampLocation_um)
        DestinationCompartmentID = BrainGenix.LibUtils.GetID.GetID(_Configuration.DestinationCompartment)
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Tool/PatchClampDAC/Create?SimulationID={_SimulationID}&Name={_Configuration.Name}&DestinationCompartmentID={DestinationCompartmentID}&ClampLocation_um={ClampLocation}")
        assert(Response != None)
        self.ID = Response["PatchClampDACID"]


    ## Access Methods
    def SetOutputList(self, _Timestep_ms:float, _VoltageList:list):
        OutputList = json.dumps(_VoltageList)
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Tool/PatchClampDAC/SetOutputList?SimulationID={self.SimulationID}&TargetDAC={self.ID}&Timestep_ms={_Timestep_ms}&DACVoltages_mV={OutputList}")
        assert(Response != None)
        return Response["StatusCode"]


# BrainGenix-NES
# AGPLv3

import requests

from . import Configuration
from . import Modes
from . import RequestHandler

from .. import Simulation


def GetTokenFromCredentials(_Username:str, _Password:str, _BaseURI:str = "https://api.braingenix.org"):

    RequestURI = f"{_BaseURI}/Auth/GetToken?Username={_Username}&Password={_Password}"
    ServerRequst = requests.get(RequestURI)
    ResponseJSON = ServerRequst.json()

    if (ResponseJSON["StatusCode"] != 0):
        raise ValueError("Invalid Credentials Passed. Please Check Your Username/Password!")
    else:
        return ResponseJSON["AuthKey"]



class Client:

    def __init__(self, _Configuration:Configuration):

        # Get And Validate Configuration
        self.Configuration = _Configuration
        self.ValidateConfig()
        self.Setup()


    # Helper Functions
    def ValidateConfig(self):

        # Check Mode
        ValidModes =  [Modes.Remote, Modes.Local]
        if self.Configuration.Mode not in ValidModes:
            raise Exception("NES Client Configuration Mode Not Set")


        
        # Local not yet implemented
        if self.Configuration.Mode == Modes.Local:
            raise NotImplementedError("Randal - please copy this over from your vbp repo.")

        # Remote Mode
        elif self.Configuration.Mode == Modes.Remote:

            # Check That User Has Actually Set Host/Port
            if (self.Configuration.Host == ""):
                raise ValueError("Please Set The API Host (ClientConfig.Host string)")
            if (self.Configuration.Port == 0):
                raise ValueError("Please Set The API Port (ClientConfig.Port int)")
            if (self.Configuration.UseHTTPS == None):
                raise ValueError("Please Indicate If You Would Like To Use HTTPS Or HTTP (ClientConfig.UseHTTPS true/false)")
            if (self.Configuration.AuthenticationMethod not in ['Token', 'Password']):
                raise ValueError("Please Set The Method Of Authentication You Are Using (ClientConfig.AuthenticationMethod 'Token'/'Password')")
            if (self.Configuration.Token == "" and self.Configuration.AuthenticationMethod == "Token"):
                raise ValueError("Please Set The API Token (ClientConfig.Token string)")
            if (self.Configuration.Username == "" and self.Configuration.AuthenticationMethod == "Password"):
                raise ValueError("Please Set Your API Username (ClientConfig.Username string)")
            if (self.Configuration.Password == "" and self.Configuration.AuthenticationMethod == "Password"):
                raise ValueError("Please Set Your API Password (ClientConfig.Password string)")


            # Build Request URI Base
            ProtocolBase = "http"
            if (self.Configuration.UseHTTPS):
                ProtocolBase = "https"
            self.URIBase = f"{ProtocolBase}://{self.Configuration.Host}:{self.Configuration.Port}"



    def Setup(self): # this should do some logic to make sure the connection is actually good.

        # If We're Operating With Password Mode, Get The Token
        if (self.Configuration.AuthenticationMethod == "Password"):
            self.Configuration.Token = GetTokenFromCredentials(self.Configuration.Username, self.Configuration.Password, self.URIBase)

        self.RequestHandler = RequestHandler.RequestHandler(self.URIBase, self.Configuration.Token)
        self.HasConnection = True


    # Getter Functions
    def IsReady(self):
        return self.HasConnection
    
    def GetAPIVersion(self):
        Status = self.RequestHandler.MakeQuery("/Diagnostic/Version")
        return Status['Version']


    # Creation Functions
    def CreateSimulation(self, _SimulationConfig):
        return Simulation.Simulation(_SimulationConfig, self.RequestHandler)



# BrainGenix-NES
# AGPLv3


# Configuration 'struct' for the client
class Configuration:

    def __init__(self):

        self.Mode = "Remote"

        self.Host:str = ""
        self.Port:int = 0
        self.UseHTTPS:bool = None

        self.AuthenticationMethod:str = "Token"
        self.Username:str = ""
        self.Password:str = ""
        self.Token:str = ""
        



# BrainGenix-NES
# AGPLv3


class Configuration():

    def __init__(self):
        
        # Create Attributes
        self.Name:str = None
        self.SourceCompartment = None
        self.ClampLocation_um:list = None
    



# BrainGenix-NES
# AGPLv3

import json

from . import Configuration

from BrainGenix.NES.Client import RequestHandler

import BrainGenix.LibUtils.GetID
import BrainGenix.LibUtils.ConfigCheck


class BSNeuron:

    def __init__(self, _Configuration:Configuration, _RequestHandler:RequestHandler, _SimulationID:int):
        # Create Attributes
        self.Name = _Configuration.Name
        self.RequestHandler = _RequestHandler

        # Run Configuration Check
        BrainGenix.LibUtils.ConfigCheck.ConfigCheck(_Configuration)

        # Create BSNeuron On Server
        SomaID = BrainGenix.LibUtils.GetID.GetID(_Configuration.SomaID)
        AxonID = BrainGenix.LibUtils.GetID.GetID(_Configuration.AxonID)
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Neuron/BSNeuron/Create"
            + f"?SimulationID={_SimulationID}"
            + f"&Name={_Configuration.Name}"
            + f"&SomaID={SomaID}"
            + f"&AxonID={AxonID}"
            + f"&MembranePotential_mV={_Configuration.MembranePotential_mV}"
            + f"&RestingPotential_mV={_Configuration.RestingPotential_mV}"
            + f"&SpikeThreshold_mV={_Configuration.SpikeThreshold_mV}"
            + f"&DecayTime_ms={_Configuration.DecayTime_ms}"
            + f"&AfterHyperpolarizationAmplitude_mV={_Configuration.AfterHyperpolarizationAmplitude_mV}"
            + f"&PostsynapticPotentialRiseTime_ms={_Configuration.PostsynapticPotentialRiseTime_ms}"
            + f"&PostsynapticPotentialDecayTime_ms={_Configuration.PostsynapticPotentialDecayTime_ms}"
            + f"&PostsynapticPotentialAmplitude_mV={_Configuration.PostsynapticPotentialAmplitude_mV}"
            )
        assert(Response != None)
        self.ID = Response["NeuronID"]



# BrainGenix-NES
# AGPLv3

import json

from . import Configuration

from BrainGenix.NES.Client import RequestHandler

import BrainGenix.LibUtils.GetID
import BrainGenix.LibUtils.ConfigCheck


class Staple:

    def __init__(self, _Configuration:Configuration, _RequestHandler:RequestHandler, _SimulationID:int):
        # Create Attributes
        self.Name =  _Configuration.Name
        self.RequestHandler = _RequestHandler

        # Run Configuration Check
        BrainGenix.LibUtils.ConfigCheck.ConfigCheck(_Configuration)

        # Create Box On Server
        SourceCompartmentID = BrainGenix.LibUtils.GetID.GetID(_Configuration.SourceCompartment)
        DestinationCompartmentID = BrainGenix.LibUtils.GetID.GetID(_Configuration.DestinationCompartment)
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Connection/Staple/Create?SimulationID={_SimulationID}&Name={_Configuration.Name}&SourceCompartmentID={SourceCompartmentID}&DestinationCompartmentID={DestinationCompartmentID}")
        assert(Response != None)
        self.ID = Response["StapleID"]


# BrainGenix-NES
# AGPLv3

from .. import Models
from . import Configuration

from BrainGenix.NES.Client import RequestHandler

from BrainGenix.NES.Shapes import Sphere
from BrainGenix.NES.Shapes import Box
from BrainGenix.NES.Shapes import Cylinder

from BrainGenix.NES.VSDA import EM
from BrainGenix.NES.Models.Connections import Staple, Receptor
from BrainGenix.NES import Tools

import BrainGenix.LibUtils.ConfigCheck


class Simulation():

    def __init__(self, _Configuration:Configuration, _RequestHandler:RequestHandler):
        # Create Attributes
        self.Name = _Configuration.Name
        self.RequestHandler = _RequestHandler

        # Run Configuration Check
        BrainGenix.LibUtils.ConfigCheck.ConfigCheck(_Configuration)

        # Create Sim On Server
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Simulation/Create?SimulationName={_Configuration.Name}")
        assert(Response != None)
        self.ID = Response["SimulationID"]


    
    ## Methods For Adding Objects

     # Tool Init Commands
    def AddPatchClampDAC(self, _PatchClampDACConfig:Tools.PatchClampDAC.Configuration):
        return Tools.PatchClampDAC.PatchClampDAC(_PatchClampDACConfig, self.RequestHandler, self.ID)

    def AddPatchClampADC(self, _PatchClampADCConfig:Tools.PatchClampADC.Configuration):
        return Tools.PatchClampADC.PatchClampADC(_PatchClampADCConfig, self.RequestHandler, self.ID)


     # Connection Init Commands
    def AddReceptor(self, _ReceptorConfig:Receptor.Configuration):
        return Receptor.Receptor(_ReceptorConfig, self.RequestHandler, self.ID)

    def AddStaple(self, _StapleConfig:Staple.Configuration):
        return Staple.Staple(_StapleConfig, self.RequestHandler, self.ID)


     # VSDA Init Commands
    def AddVSDAEM(self, _VSDAEMConfig:EM.Configuration):
        return EM.EM(_VSDAEMConfig, self.RequestHandler, self.ID)


     # Compartments Add Methods
    def AddBSCompartment(self, _BSCompartmentConfig:Models.Compartments.BS.Configuration):
        return Models.Compartments.BS.BS(_BSCompartmentConfig, self.RequestHandler, self.ID)

    # Neurons Add Methods
    def AddBSNeuron(self, _BSNeuronConfig:Models.Neurons.BS.Configuration):
        return Models.Neurons.BS.BSNeuron(_BSNeuronConfig, self.RequestHandler, self.ID)

    
     # Geometry Add Methods
    def AddSphere(self, _SphereConfig:Sphere.Configuration):
        return Sphere.Sphere(_SphereConfig, self.RequestHandler, self.ID)

    def AddBox(self, _BoxConfig:Box.Configuration):
        return Box.Box(_BoxConfig, self.RequestHandler, self.ID)

    def AddCylinder(self, _CylinderConfig:Cylinder.Configuration):
        return Cylinder.Cylinder(_CylinderConfig, self.RequestHandler, self.ID)


    ## Simulation Update Routes
    def Reset(self):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Simulation/Reset?SimulationID={self.ID}")
        assert(Response != None)
        return Response

    def RunFor(self, _SimulationDuration_ms:float):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Simulation/RunFor?SimulationID={self.ID}&Runtime_ms={_SimulationDuration_ms}")
        assert(Response != None)
        return Response

    def RecordAll(self, _MaxRecordTime_ms:float):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Simulation/RecordAll?SimulationID={self.ID}&MaxRecordTime_ms={_MaxRecordTime_ms}")
        assert(Response != None)
        return Response

    def GetRecording(self):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Simulation/GetRecording?SimulationID={self.ID}")
        assert(Response != None)
        print("GetRecording Is Not Yet Fully Implemented (FYI)!")
        print(Response)
        return Response

    def GetStatus(self):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Simulation/GetStatus?SimulationID={self.ID}")
        assert(Response != None)
        return Response



# BrainGenix-NES
# AGPLv3

import json

from . import Configuration

from BrainGenix.NES.Client import RequestHandler

import BrainGenix.LibUtils.ConfigCheck


class Box:

    def __init__(self, _Configuration:Configuration, _RequestHandler:RequestHandler, _SimulationID:int):
        # Create Attributes
        self.Name = _Configuration.Name
        self.RequestHandler = _RequestHandler

        # Run Configuration Check
        BrainGenix.LibUtils.ConfigCheck.ConfigCheck(_Configuration)

        # Create Box On Server
        CenterPos = json.dumps(_Configuration.CenterPosition_um)
        Dimensions = json.dumps(_Configuration.Dimensions_um)
        Rotation = json.dumps(_Configuration.Rotation_rad)
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Geometry/Shape/Box/Create?SimulationID={_SimulationID}&Name={_Configuration.Name}&CenterPosition_um={CenterPos}&Dimensions_um={Dimensions}&Rotation_rad={Rotation}")
        assert(Response != None)
        self.ID = Response["ShapeID"]


# BrainGenix-NES
# AGPLv3

import inspect


def ConfigCheck(_ConfigurationObject:object):

    # Firstly, we're going to get a list of all attributes of this object
    AttributeList = dir(_ConfigurationObject)

    # Now, we enumerate these attribs, filter out those builtin to python (starting with __)
    for AttributeName in AttributeList:

        # If it starts with an underscore, skip it
        if (AttributeName.startswith("__")):
            continue

        # Now, check the value of this attribute
        ThisAttribute = getattr(_ConfigurationObject, AttributeName)

        # Check the values of these based on type, it should *never* be empty (we want users to fill out the config objects fully)
        if (ThisAttribute == ""):
            PrintErrorMessage(_ConfigurationObject, AttributeName)
        elif (ThisAttribute == None):
            PrintErrorMessage(_ConfigurationObject, AttributeName)



def PrintErrorMessage(_ConfigurationObject:object, _AttributeName:str):
    
    # Get Information From Class To Build Error Message
    ClassFileName:str = inspect.getfile(_ConfigurationObject.__class__).split("BrainGenix/")[1].strip(".py").replace("/", ".")
    ErrorMessage:str = f'''

    -----------------------
    Configuration Error In '{ClassFileName}', Attribute '{_AttributeName}' Was Left Uninitialized.
    Please Reference The API Spec For Usage Information.
    https://gitlab.braingenix.org/carboncopies/BrainGenix-API/-/blob/master/Docs/API.md?ref_type=heads
    If You Feel That This Is A Bug, Please Open An Issue Here:
    https://gitlab.braingenix.org/carboncopies/BrainGenix/NES/PythonClient/-/issues
    -----------------------
    '''

    # Now, Chuck An Error To stderr Alerting The User That There Is A Config Issue
    raise AttributeError(ErrorMessage)


# BrainGenix-NES
# AGPLv3

import json

from . import Configuration

from BrainGenix.NES.Client import RequestHandler
import BrainGenix.LibUtils.GetID

import BrainGenix.LibUtils.ConfigCheck


class PatchClampADC:

    def __init__(self, _Configuration:Configuration, _RequestHandler:RequestHandler, _SimulationID:int):
        # Create Attributes
        self.Name =  _Configuration.Name
        self.RequestHandler = _RequestHandler
        self.SimulationID = _SimulationID

        # Run Configuration Check
        BrainGenix.LibUtils.ConfigCheck.ConfigCheck(_Configuration)

        # Create On Server
        ClampLocation = json.dumps(_Configuration.ClampLocation_um)
        SourceCompartmentID = BrainGenix.LibUtils.GetID.GetID(_Configuration.SourceCompartment)
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Tool/PatchClampADC/Create?SimulationID={_SimulationID}&Name={_Configuration.Name}&SourceCompartmentID={SourceCompartmentID}&ClampLocation_um={ClampLocation}")
        assert(Response != None)
        self.ID = Response["PatchClampADCID"]


    ## Access Methods
    def SetSampleRate(self, _Timestep_ms:float):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Tool/PatchClampADC/SetSampleRate?SimulationID={self.SimulationID}&TargetADC={self.ID}&Timestep_ms={_Timestep_ms}")
        assert(Response != None)
        return Response["StatusCode"]

    def GetRecordedData(self):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/Tool/PatchClampADC/GetRecordedData?SimulationID={self.SimulationID}&TargetADC={self.ID}")
        assert(Response != None)
        return Response


# BrainGenix-NES
# AGPLv3


class Configuration():

    def __init__(self):
        
        # Create Attributes
        self.PixelResolution_nm:float = None
        self.ImageWidth_px:int = None
        self.ImageHeight_px:int = None
        self.SliceThickness_nm:float = None
        self.ScanRegionOverlap_percent:float = None 
        self.MicroscopeFOV_deg:float = None
        self.NumPixelsPerVoxel_px:int = None
    



# BrainGenix-NES
# AGPLv3

import json
import base64
import tqdm
import yaspin
import time

from . import Configuration

from BrainGenix.NES.Client import RequestHandler

import BrainGenix.LibUtils.ConfigCheck



class EM:

    def __init__(self, _Configuration:Configuration, _RequestHandler:RequestHandler, _SimulationID:int):
        # Create Attributes
        self.RequestHandler = _RequestHandler
        self.SimulationID = _SimulationID

        # Run Configuration Check
        BrainGenix.LibUtils.ConfigCheck.ConfigCheck(_Configuration)

        # Create On Server
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/VSDA/EM/Initialize?SimulationID={_SimulationID}")
        assert(Response != None)

        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/VSDA/EM/SetupMicroscope?SimulationID={_SimulationID}&PixelResolution_nm={_Configuration.PixelResolution_nm}&ImageWidth_px={_Configuration.ImageWidth_px}&ImageHeight_px={_Configuration.ImageHeight_px}&SliceThickness_nm={_Configuration.SliceThickness_nm}&ScanRegionOverlap_percent={_Configuration.ScanRegionOverlap_percent}&MicroscopeFOV_deg={_Configuration.MicroscopeFOV_deg}&NumPixelsPerVoxel_px={_Configuration.NumPixelsPerVoxel_px}")
        assert(Response != None)



    ## Access Methods
    def DefineScanRegion(self, _Point1_um:list, _Point2_um:list):
        Point1 = json.dumps(_Point1_um)
        Point2 = json.dumps(_Point2_um)
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/VSDA/EM/DefineScanRegion?SimulationID={self.SimulationID}&Point1_um={Point1}&Point2_um={Point2}")
        assert(Response != None)
        self.ScanRegionID = Response["ScanRegionID"]


    def QueueRenderOperation(self):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/VSDA/EM/QueueRenderOperation?SimulationID={self.SimulationID}&ScanRegionID={self.ScanRegionID}")
        assert(Response != None)
        return Response["StatusCode"]


    def GetRenderStatus(self):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/VSDA/EM/GetRenderStatus?SimulationID={self.SimulationID}")
        assert(Response != None)
        return Response
    

    def GetImageStack(self):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/VSDA/EM/GetImageStack?SimulationID={self.SimulationID}&ScanRegionID={self.ScanRegionID}")
        assert(Response != None)
        return Response["RenderedImages"]


    def GetImage(self, _ImageHandle:str):
        Response = self.RequestHandler.MakeAuthenticatedQuery(f"/NES/VSDA/EM/GetImage?SimulationID={self.SimulationID}&ImageHandle={_ImageHandle}")
        assert(Response != None)
        return bytes(Response["ImageData"], 'utf-8')
    

    def WaitForRender(self):

        # Setup Status Information
        StatusInfo:dict = self.GetRenderStatus()


        # Perform Sanity Check On Render
        if (StatusInfo["RenderStatus"] < 3):
            print("Error during rendering, API reports:\n")
            print(StatusInfo)


        # Block With Queued Spinner
        QueueSpinner = yaspin.yaspin(text="Render Operation In Queue, Elapsed Time", color="green", timer=True)
        QueueSpinner.start()
        while (StatusInfo["RenderStatus"] == 3):

            # Get Status Info, Wait To Avoid API Spam
            StatusInfo:dict = self.GetRenderStatus()
            time.sleep(0.25)

        QueueSpinner.ok()


        
        # Setup Slice, Image Status Bar
        SliceStatusBar = tqdm.tqdm("Rendering Slice", total=1)
        SliceStatusBar.leave = True
        SliceStatusBar.colour = "green"
        LastSliceNumber = 0

        ImageStatusBar = tqdm.tqdm("Rendering Image", total=1)
        ImageStatusBar.leave = True
        ImageStatusBar.colour = "magenta"


        # Block Execution Until Render Finishes, Update Bar As We Wait
        while (StatusInfo["RenderStatus"] != 5):

            # Update Slice Bar
            SliceStatusBar.total = int(StatusInfo["TotalSlices"])
            SliceStatusBar.n = max(int(StatusInfo["CurrentSlice"]), 0)
            SliceStatusBar.refresh()
            SliceStatusBar.set_description(f"Rendering Slice {StatusInfo['CurrentSlice']}/{StatusInfo['TotalSlices']}")


            # Update Image Bar
            if (SliceStatusBar.n > LastSliceNumber):
                LastSliceNumber = SliceStatusBar.n
            ImageStatusBar.total = int(StatusInfo["TotalSliceImages"])
            ImageStatusBar.n = max(int(StatusInfo["CurrentSliceImage"]), 0)
            ImageStatusBar.refresh()
            ImageStatusBar.set_description(f"Rendering Image {StatusInfo['CurrentSliceImage']}/{StatusInfo['TotalSliceImages']}")



            # Get Status Info, Wait To Avoid API Spam
            StatusInfo:dict = self.GetRenderStatus()
            time.sleep(0.1)
        
    
        # When we're done, rendering is finished - make it look like it truly is, then close the bar
        SliceStatusBar.total = int(StatusInfo["TotalSlices"])
        SliceStatusBar.n = int(StatusInfo["TotalSlices"])
        SliceStatusBar.set_description(f"Rendering Slice {StatusInfo['TotalSlices']}/{StatusInfo['TotalSlices']}")
        SliceStatusBar.refresh()
        SliceStatusBar.close()
        ImageStatusBar.total = int(StatusInfo["TotalSliceImages"])
        ImageStatusBar.n = int(StatusInfo["TotalSliceImages"])
        ImageStatusBar.set_description(f"Rendering Image {StatusInfo['TotalSliceImages']}/{StatusInfo['TotalSliceImages']}")
        ImageStatusBar.refresh()
        ImageStatusBar.close()



    def SaveImageStack(self, _ImageStackDirectoryPrefix:str = ""):

        # Get Image Stack Manifest
        ImageHandles = self.GetImageStack()

        # Setup Progress Bar
        Bar = tqdm.tqdm("Downloading Image Stack", total=len(ImageHandles))
        Bar.leave = True
        Bar.colour = "green"


        for i in range(len(ImageHandles)):

            # Save This Image
            ImageData = self.GetImage(ImageHandles[i])
            with open(_ImageStackDirectoryPrefix + ImageHandles[i].split("/")[1], "wb") as FileHandler:
                FileHandler.write(base64.decodebytes(ImageData))

            Bar.set_description(f"Downloading Image {i + 1}/{len(ImageHandles)}")

            # Count Up Bar
            Bar.n = i + 1
            Bar.refresh()

        Bar.close()


# BrainGenix-NES
# AGPLv3

import requests

class RequestHandler:

    def __init__(self, _URIBase:str, _Token:str):
        self.URIBase = _URIBase
        self.Token = _Token

        # Check that API is up
        try:
            Response = requests.get(f"{self.URIBase}/Hello")
        except:
            raise ConnectionError("Unable To Connect To API Endpoint")
        

    
    # Make Query
    def MakeQuery(self, _URIStub:str):
        
        # Exception Handling For Network Errors
        Response = None
        for RetryCount in range(5):
            try:
                Response = requests.get(f"{self.URIBase}{_URIStub}")
                break
            except requests.exceptions.ChunkedEncodingError as E:
                if (RetryCount == 4):
                    raise ConnectionError("Error Communicating With API, An IncompleteRead Has Occured!")
                else:
                    print("\n---------------")
                    print(f"Warning, Attempt {RetryCount} Has Failed, Retrying Query {self.URIBase}{_URIStub}")
                    print(f"Python Reports Error: {E}")
                    print("---------------\n")

        # Parse the Response, Return It
        ResponseJSON = Response.json()
        if (ResponseJSON["StatusCode"] != 0):
            raise ConnectionError(f"Error During API Call To '{self.URIBase}{_URIStub}', API Returned Status Code '{ResponseJSON['StatusCode']}'")
            return None
        return ResponseJSON


    def MakeAuthenticatedQuery(self, _URIStub:str):
        URIStub = f"{_URIStub}&AuthKey={self.Token}"
        return self.MakeQuery(URIStub)


# BrainGenix-NES
# AGPLv3


class Configuration():

    def __init__(self):
        
        # Create Attributes
        self.Name:str = None
        self.CenterPosition_um:list = None
        self.Dimensions_um:list = None
        self.Rotation_rad:list = None
    



# BrainGenix-NES
# AGPLv3


class Configuration():

    def __init__(self):
        
        # Create Attributes
        self.Name:str = None
        self.SomaID:str = None
        self.AxonID:str = None
        self.MembranePotential_mV:float = None
        self.RestingPotential_mV:float = None
        self.SpikeThreshold_mV:float = None
        self.DecayTime_ms:float = None
        self.AfterHyperpolarizationAmplitude_mV:float = None
        self.PostsynapticPotentialRiseTime_ms:float = None
        self.PostsynapticPotentialDecayTime_ms:float = None
        self.PostsynapticPotentialAmplitude_mV:float = None
